/*
** safeio.c - Signal/Async safe wrapper functions
**
** Copyright (c) 1997-1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <sys/mman.h>
#ifndef MAP_FAILED
#define MAP_FAILED -1
#endif

#include "pftpd.h"


void
s_abort(void)
{
    int *p = (int *) NULL;

    *p = 4711;
    abort();
}



int
s_open(const char *path,
       int oflag,
       ...)
{
    int s;
    mode_t mode = 0;

    if (oflag & O_CREAT)
    {
	va_list ap;

	va_start(ap, oflag);
	/* FIXME: need to use widened form of mode_t here.  */
	mode = va_arg(ap, int);
	va_end(ap);
    }

    while ((s = open(path, oflag, mode)) < 0 && errno == EINTR)
	;
    
    if (s < 0 && (errno == EMFILE
		  || errno == ENFILE
		  || errno == ENOMEM 
#ifdef ENOSR
		  || errno == ENOSR
#endif
		  ))
    {
	/* Too many open files */
	
	syslog(LOG_WARNING, "s_open(\"%s\", 0%o): %m", path, oflag);
    }
    
    return s;
}



ssize_t
s_write(int fd,
	const void *buf,
	size_t len)
{
    ssize_t code;
    
    while ((code = write(fd, buf, len)) < 0 && errno == EINTR)
	;

    return code;
}



ssize_t
s_read(int fd,
       void *buf,
       size_t len)
{
    ssize_t code;

    
    while ((code = read(fd, buf, len)) < 0 && errno == EINTR)
	;
    
    return code;
}



ssize_t
s_pread(int fd,
	void *buf,
	size_t len,
	off_t off)
{
    ssize_t code;
    
#ifdef HAVE_PREAD
    while ((code = pread(fd, buf, len, off)) < 0 && errno == EINTR)
	;
    return code;
#else
    if (lseek(fd, off, SEEK_SET) == (off_t) -1)
	return -1;
    
    while ((code = read(fd, buf, len)) < 0 && errno == EINTR)
	;
    return code;
#endif
}


int
s_close(int fd)
{
    int code;
    
    while ((code = close(fd)) < 0 && errno == EINTR)
	;
    
    return code;
}



/*
** A "safe" malloc, that always succeeds (or logs an
** error to syslog and the abort()'s.
**
** The buffer returned is zeroed out.
*/
void *
s_malloc(size_t size)
{
    void *p;

    p = (void *) malloc(size);
    if (p == NULL)
    {
	if (debug)
	    fprintf(stderr, "s_malloc(%lu) failed - aborting\n",
		    (unsigned long) size);
	
	syslog(LOG_ERR, "malloc(%lu): %m", (unsigned long) size);
	s_abort();
    }

    memset(p, 0, size);
    return p;
}


void *
s_realloc(void *oldp, size_t size)
{
    void *p;

    
    if (oldp)
	p = (void *) realloc(oldp, size);
    else
	p = (void *) malloc(size);
    
    if (p == NULL)
    {
	if (debug)
	    fprintf(stderr, "s_realloc(%lu) failed - aborting\n",
		    (unsigned long) size);
	
	syslog(LOG_ERR, "realloc(%lu): %m", (unsigned long) size);
	s_abort();
    }

    return p;
}


void
s_free(void *p)
{
    if (p != NULL)
	free(p);
}


char *
s_strdup(const char *s)
{
    char *ns;
    size_t len;
    
    
    if (s == NULL)
	return NULL;

    len = strlen(s)+1;
    ns = (char *) malloc(len);
    if (ns == NULL)
    {
	syslog(LOG_ERR, "strdup(): malloc(%lu): %m", (unsigned long) len);
	s_abort();
    }

    memcpy(ns, s, len);
    return ns;
}


int
s_accept(int fd,
	 struct sockaddr *sin,
	 socklen_t *len)
{
    int new_fd;


    while ((new_fd = accept(fd, sin, len)) < 0 && errno == EINTR)
	;

    return new_fd;
}




static pthread_mutex_t random_lock;
static pthread_once_t random_once = PTHREAD_ONCE_INIT;

static void
random_lock_init(void)
{
    unsigned int seed;
    
    pthread_mutex_init(&random_lock, NULL);
    
    seed = time(NULL);
#ifdef HAVE_SRANDOM
    srandom(seed);
#else
    srand(seed);
#endif
}


long
s_random(void)
{
    long res;
    
    pthread_once(&random_once, random_lock_init);
    
    pthread_mutex_lock(&random_lock);
#ifdef HAVE_RANDOM
    res = random();
#else
    res = rand();
#endif
    pthread_mutex_unlock(&random_lock);

    return res;
}



int
s_vsnprintf(char *buf,
	    size_t bufsize,
	    const char *format,
	    va_list ap)
{
    int retcode;


    if (bufsize < 1)
    {
	if (debug)
	    fprintf(stderr, "s_snprintf(..., %d, ...): illegal bufsize\n",
		    bufsize);
	syslog(LOG_ERR, "s_snprintf(..., %d, ...): illegal bufsize",
	       bufsize);
	s_abort();
    }
    
    buf[bufsize-1] = '\0';
#ifdef HAVE_VSNPRINTF
    retcode = vsnprintf(buf, bufsize, format, ap);
#else
#ifdef SPRINTF_RETVAL_POINTER
    /* XXX: The reason we check for sprintf()'s return type and not
       vsprintf() (which we should) is that SunOS 4 doesn't declare
       vsprintf() in any header files, but it does have the same return
       value as sprintf(). So expect a compiler warning here unless
       we use the (uncesserary) '(char *)' cast - *Sigh* */
    {
	char *cp = (char *) vsprintf(buf, format, ap);
	if (cp == NULL)
	    retcode = -1;
	else
	    retcode = strlen(cp);
    }
#else
    retcode = vsprintf(buf, format, ap);
#endif
#endif
    if (debug > 3)
	fprintf(stderr, "s_snprintf(%08lx, %d, \"%s\", ...) = %d\n",
		(unsigned long) buf, bufsize, format, retcode);
    
    if (retcode > 0 && (buf[bufsize-1] != '\0' ||
			retcode > bufsize-1))
    {
	if (debug)
	    fprintf(stderr, "s_snprintf(..., %d, ...) = %d: buffer overrun\n",
		    bufsize, retcode);
	syslog(LOG_ERR, "s_snprintf(..., %d, ...) = %d: buffer overrun\n",
	       bufsize, retcode);
	
	s_abort();
    }

    return retcode;
}



int
s_snprintf(char *buf,
	    size_t bufsize,
	    const char *format,
	    ...)
{
    va_list ap;
    int retcode;
    

    va_start(ap, format);
    retcode = s_vsnprintf(buf, bufsize, format, ap);
    va_end(ap);

    return retcode;
}



/*
** Set up a signal handler without restartable syscall.
*/
SIGHANDLER_T
s_signal(int sig,
	 SIGHANDLER_T handler)
{
#ifdef HAVE_SIGACTION
    if (handler != SIG_IGN)
    {
	struct sigaction osa, nsa;
	
	if (sigaction(sig, NULL, &osa) == -1)
	    return SIG_ERR;

	nsa = osa;
	nsa.sa_handler = handler;
	nsa.sa_flags = 0;
	sigaction(sig, &nsa, NULL);
	return osa.sa_handler;
    }
#endif
    return signal(sig, handler);
}


#define COPYFD_BUFSIZE (64*1024)

static int
simple_copyfd(int to_fd,
	      int from_fd,
	      off_t *start,
	      size_t maxlen)
{
    char buf[COPYFD_BUFSIZE];
    int err, tlen, written, i;


    if (debug)
	fprintf(stderr, "simply_copyfd(%d, %d, %lu, %d)\n",
		to_fd, from_fd, (start != NULL ? *start : 0), maxlen);
    
    if (start && *start > 0)
	if (lseek(from_fd, *start, SEEK_SET) == (off_t) -1)
	{
	    /* Can't seek - wind to the starting point */
	    tlen = 0;
	    while (tlen < *start)
	    {
		i = *start - tlen;
		if (i > sizeof(buf))
		    i = sizeof(buf);
		
		err = s_read(from_fd, buf, i);
		if (err <= 0)
		    return -1;
		
		tlen += err;
	    }
	    
	    if (tlen < *start)
		return -1;		
	}
    
    written = 0;
    tlen = -1;
    
    while ((maxlen == 0 || written < maxlen) &&
	   (tlen = s_read(from_fd, buf,
			  (maxlen == 0 || maxlen - written > sizeof(buf) ?
			   sizeof(buf) : maxlen - written))) > 0)
    {
	err = s_write(to_fd, buf, tlen);
	if (err < 0)
	    return -1;
	
	written += tlen;
	if (start)
	    *start += tlen;
    }
    
    if (tlen < 0)
	return -1;

    return written;
}


int
s_copyfd(int to_fd,
	 int from_fd,
	 off_t *start,
	 size_t maxlen)
{
    struct stat sb;
    char *buf;
    int err, len;
    

    /* If source isn't a plain file, use simple copyfd */
    if (fstat(from_fd, &sb) < 0)
	return simple_copyfd(to_fd, from_fd, start, maxlen);
    
    if (start && (*start >= sb.st_size))
	return -1;

    if (maxlen == 0)
	len = sb.st_size-(start != NULL ? *start : 0);
    else
	len = maxlen;

#ifdef HAVE_SENDFILE
    /* Assuming Linux style sendfile() */
    while ((err = sendfile(to_fd, from_fd, start, len)) == -1 &&
	   errno == EINTR)
	;

#if 0 /* Enable when we know what Exxx Linux returns for invalid fd's */
    if (err == -1 && errno == Exxxx)
	return simple_copyfd(to_fd, from_fd, start, maxlen);
#endif
    
    return err;
#else
    if (sb.st_size <= COPYFD_BUFSIZE)
	return simple_copyfd(to_fd, from_fd, start, maxlen);

    
    buf = mmap(NULL, sb.st_size, PROT_READ, MAP_PRIVATE,
	       from_fd, 0);
    if (buf == MAP_FAILED)
	return simple_copyfd(to_fd, from_fd, start, maxlen);

#ifdef MADV_SEQUENTIAL
    /* XXX: Enabled MADV_WILLNEED? */
    madvise(buf, sb.st_size, MADV_SEQUENTIAL /* | MADV_WILLNEED */ );
#endif

    /* XXX: Lock the file region? */
    err = s_write(to_fd, buf+(start != NULL ? *start : 0), len);
    munmap(buf, sb.st_size);
    if (err > 0 && start)
	*start += err;
#endif
    
    return err;
}


int
s_bind(int fd,
       struct sockaddr *sin,
       socklen_t len)
{
    int err;


    while ((err = bind(fd, sin, len)) < 0 && errno == EINTR)
	;

    return err;
}

int
s_connect(int fd,
	  struct sockaddr *sin,
	  socklen_t len)
{
    int err;


    while ((err = connect(fd, sin, len)) < 0 && errno == EINTR)
	;

    return err;
}


char *
s_inet_ntoa(struct in_addr ia,
	    char *buf,
	    size_t bufsize)
{
    UINT32 addr;

    
    if (bufsize < 16)
	return NULL;

    addr = ntohl(ia.s_addr);
    s_snprintf(buf, bufsize, "%u.%u.%u.%u",
	       (addr >> 24) & 0xFF,
	       (addr >> 16) & 0xFF,
	       (addr >>  8) & 0xFF,
	       (addr      ) & 0xFF);

    return buf;
}
